//    This file is part of RoadColoringImpl.
//
//    Copyright (C) 2008 Florian SIKORA sikora@univ-mlv.fr
//    http://igm.univ-mlv.fr/~fsikora/index.php?p=r
//
//    RoadColoringImpl is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    RoadColoringImpl is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    The complete GNU General Public Licence Notice can be found as the
//    `COPYING' file in the root directory.


package fr.umlv.roadcoloring.graph;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.Arrays;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;
import java.util.Scanner;

/**
 * 
 * Manage a graph strongly connected, aperiodic, 2 colored (blue and red)
 * run coloring() for change the coloring in order to have a synchronized word
 * 
 * Can be generated randomly if size n, or from a file
 * The syntax of the graph is :
 * n
 * i r b
 * 
 * with n size of the graph
 * and, for each node i, the red edge and the blue edge
 * 
 * Can be seen by dotty()
 * 
 * @author fsikora
 *
 */

public class Graph {
	//edges
	private int[] red;
	private int[] blue;
	//size
	private int n = 0;

	protected final static byte REDEDGE = 0;
	protected final static byte BLUEEDGE = 1;

	private int[] classes;

	/**
	 * Generate a randomly acceptable graph of size n
	 * @param n	size of graph
	 * @param generatorPgrm	a program for generate a randomly graph. If null, use Random JDK classes
	 * @throws IOException
	 */
	public Graph(int n, String generatorPgrm) throws IOException {
		this.n = n;
		red = new int[n];
		blue = new int[n];
		classes = new int[n];

		//generate a graph while is not acceptable
		//	do {
		generator(n, generatorPgrm);
		//	}
		//	while(!isAcceptable());
	}

	/**
	 * Create a graph from another graph
	 * create array but not the edges
	 * @param g
	 */
	protected Graph(Graph g) {
		this.n = g.getSize();
		this.red = new int[this.n];
		this.blue = new int[this.n];
		/*
		for (int i=0 ; i<this.n ; ++i) {
			this.red[i] = g.getRed(i);
			this.blue[i] = g.getBlue(i);
		}
		 */
	}


	/**
	 * Create a graph from a file
	 * @param fileName name of the graph
	 * @exception {@link IllegalArgumentException} if the graph is not acceptable
	 */
	public Graph(String fileName) {
		parsingFile(fileName);

		if (!isStronglyConnected()) {
			throw new IllegalArgumentException("Input graph " + fileName + " isn't strongly connected");
		}
		if (!isAperiodic()) {
			throw new IllegalArgumentException("Input graph " + fileName + " isn't strongly connected");
		}
	}

	/**
	 * Generate a random graph (can be not acceptable)
	 * @param n	size wanted
	 * @param generatorPgrm	a optionnal generator (can be null) who can be exec by Runtime.exec()
	 * @throws IOException
	 */
	private void generator(int n, String generatorPgrm) throws IOException {
		if (generatorPgrm == null) {
			generatorRandom(n);
			return;
		}

		String[] cmd = {generatorPgrm, String.valueOf(n)};
		InputStream is = Runtime.getRuntime().exec(cmd).getInputStream();
		//InputStream is = Runtime.getRuntime().exec("echo toto").getInputStream();		
		BufferedReader br = new BufferedReader(new InputStreamReader(is));
		parsingFile(br);
		br.close();

	}

	/**
	 * generate a random graph using Random.nextInt() for each edge
	 * @param n size wanted
	 */
	private void generatorRandom(int n) {
		Random r = new Random();

		for(int i=0 ; i<n ; ++i) {
			red[i] = r.nextInt(n);
			blue[i] = r.nextInt(n);
			//classes[i] = i;
			classes[i] = -1;
		}
	}

	/**
	 * get graph size
	 * @return graph size
	 */
	public int getSize() {
		return n;
	}

	/**
	 * get the next by red for index
	 * @param index node
	 * @return the next of index by red
	 */
	public int getRed(int index) {
		//return red.get(index);
		return red[index];
	}

	/**
	 * get the next by blue for index
	 * @param index node
	 * @return the next of index by blue
	 */
	public int getBlue(int index) {
		return blue[index];
	}

	/**
	 * Get the next by color of index
	 * @param index node
	 * @param color can by BLUEEDGE or REDEDGE
	 * @return the next by color
	 */
	public int get(int index, byte color) {
		if(color == BLUEEDGE) {
			return getBlue(index);
		}
		else if (color == REDEDGE) {
			return getRed(index);
		}
		else {
			throw new IllegalArgumentException(color + " unknown");
		}
	}

	/**
	 * set the next by red of index
	 * @param index start node
	 * @param value node next by red
	 */
	protected void setRed(int index, int value) {
		red[index] = value;
	}
	/**
	 * set the next by blue of index
	 * @param index start node
	 * @param value node next by blue
	 */
	protected void setBlue(int index, int value) {
		blue[index] = value;
	}

	/**
	 * get the classes for each node
	 * @return the array of classes
	 */
	public int[] getClasses() {
		return classes;
	}

	/**
	 * get the class of a node from the initial graph
	 * -1 is the status of a node not in a class (so return herself)
	 * @param node node which we seach class
	 * @return the node's class
	 */
	protected int getNodeClass(int node) {
		//System.out.println(classes);
		int cl = classes[node];
		if(cl == -1) return node;
		return cl;
	}

	/**
	 * test if this graph is strongly connected
	 * @return true if it's, false else
	 */
	public boolean isStronglyConnected() {
		//get the components for this graph (an array who tell on each component is attach each node)
		Integer[] components = new StronglyConnectedComponents(this).components();

		int componentNumber = components[0];
		for (int component : components) {
			if (component != componentNumber) return false;
		}
		return true;
	}

	/**
	 * tell if the graph is aperiodic
	 * @return true if it is, false else
	 */
	public boolean isAperiodic() {
		//p is the period of this graph
		int p = new Aperiodic(this).aperiodic();
		return p == 1;
	}

	/**
	 * tell if the graph is acceptable (strongly connected and aperiodic)
	 * @return true if it's acceptable, false else
	 */
	public boolean isAcceptable() {
		return isStronglyConnected() && isAperiodic();
	}

	/**
	 * check if s start with an integer
	 * @param s a string
	 * @return true is start with an integer, false else
	 */
	private boolean isInteger(String s) {
		try {
			Integer.parseInt(s.split(" ")[0]);
		}
		catch(NumberFormatException e) {
			return false;
		}
		return true;
	}


	/**
	 * parse and fill the graph from the buffered reader
	 * @param br buffered reader for read the flow
	 */
	private void parsingFile(BufferedReader br) {
		String fileUsage = "File usage : node red blue";
		try {
			String oneLine;
			Scanner scanner;

			//check the graph size
			while((oneLine = br.readLine()) != null) {
				if(oneLine.length() > 0 && oneLine.charAt(0) != '#' && oneLine.charAt(0) != '-') {
					//size can be juste n (my files) or MaxSize = n (regal files)
					if (isInteger(oneLine)) {
						scanner = new Scanner(oneLine);
					}
					else if (oneLine.startsWith("MaxSize = ")){
						scanner = new Scanner(oneLine.substring("MaxSize = ".length()));
					}
					else {
						continue;
					}

					this.n = scanner.nextInt();
					//System.out.println("n = " + n);
					red = new int[n];
					blue = new int[n];
					classes = new int[n];
					break;
				}
			}

			if (n == 0) throw new IllegalArgumentException("Size of graph is 0 !");

			//check the nodes
			while((oneLine = br.readLine()) != null) {
				if(oneLine.length() > 0 && oneLine.charAt(0) != '#' && oneLine.charAt(0) != '-') {
					//System.out.println("read " + oneLine);
					scanner = new Scanner(oneLine);
					if(!scanner.hasNextInt()) {	//ignore lines (regal case..)
						//System.out.println("ignore");
						continue;
						//throw new IllegalArgumentException(fileUsage + " (" + oneLine + ")");
					}
					int node = scanner.nextInt();
					if(!scanner.hasNextInt()) {
						throw new IllegalArgumentException(fileUsage + " (" + oneLine + ")");
					}
					int r = scanner.nextInt();
					if(!scanner.hasNextInt()) {
						throw new IllegalArgumentException(fileUsage + " (" + oneLine + ")");
					}
					int b = scanner.nextInt();

					if (node >= n || r >= n || b >= n) {
						throw new IllegalArgumentException(fileUsage + " (" + oneLine + ") (must be < to " + n);
					}

					/*red.add(node, r);
					blue.add(node, b);*/
					red[node] = r;
					blue[node] = b;
					//classes[node] = node;
					classes[node] = -1;
				}
			}

		} catch (IOException e) {
			System.err.println("IO Exception in file reader");
			e.printStackTrace();
		}
	}

	/**
	 * fill the graph from a file
	 * @param fileName name of the file
	 */
	private void parsingFile(String fileName) {
		try {
			BufferedReader br = new BufferedReader(new FileReader(fileName));
			parsingFile(br);
			br.close();
		} catch (FileNotFoundException e) {
			System.err.println(fileName + " not found");
			e.printStackTrace();
		} catch (IOException e) {
			System.err.println("Error durring closing flow");
			e.printStackTrace();
		}
	}

	/**
	 * eventually change the colloring of the graph in order to be synchronizing
	 * compute the quotient graph until it's a size > 1
	 * compute the quotient from the classes. Find two node (and more) with same class by stable pair.
	 * Find stable pair by computing levels of each node, and perform some exchange if needed
	 * 
	 * at the end of this function, the graph is seems to be synchronizing (a word who synchronized can be found)
	 */
	public void coloring() {
		QuotientGraph quotient = new QuotientGraph(this);
		int i=0;
		System.out.println("Coloring..");

		while(quotient.getQuotientSize() > 1) {
			System.out.println("\n Step " + ++i + ": ");
			System.out.println("   quotient : \n" + quotient);
			StablePair stable = quotient.findStablePair();	//get a stable pair

			if (stable != null) {	//if null, must be recompute with the new coloring
				this.updateColoring(quotient);	//check if some change in coloring must be update in the initial graph

				//the two node in stable pair are in the same class
				int oldBClass = getNodeClass(stable.getB());
				this.classes[stable.getB()] = stable.getA();
				this.classes[stable.getA()] = stable.getA();

				updateClass(oldBClass, stable.getA());

				//check if other nodes can be in same class
				System.out.println("  Extend classes..");
				this.extendClasses(stable);

				//compute a new quotient graph from the new classes
				quotient = new QuotientGraph(this);
			}
		}
	}

	/**
	 * flip the edges of the node
	 * @param node node to be flip
	 */
	protected void edgeSwap(int node) {
		int tmp = blue[node];
		blue[node] = red[node];
		red[node] = tmp;		
	}

	/**
	 * browse the red edge len time from the start node
	 * @param start start node
	 * @param len length in the red edges
	 * @return the end node
	 */
	protected int getSucessorByRed(int start, int len) {
		int iter = start;
		for(int i=0 ; i<len ; ++i) {
			iter = getRed(iter);
		}
		return iter;
	}




	/**
	 * compare red out edge of init graph and quotient graph
	 * flip edge of node in the init graph if the sucessors aren't compatible
	 * 
	 * @param quotient quotient graph
	 */
	private void updateColoring(QuotientGraph quotient) {
		for(int i=0 ; i<getSize() ; ++i) {
			//get the target class in init graph
			//int res = classes[getRed(i)];
			//if (res == -1) res = getRed(i);
			int res = getNodeClass(getRed(i));

			//find the class of i
			//int cl = classes[i];
			//if (cl == -1) cl = i;
			int cl = getNodeClass(i);

			if (res != quotient.getRed(cl)) {
				System.out.println("\texchange edge of " + i);
				edgeSwap(i);
			}
		}
	}

	/**
	 * Get all nodes from a class (browse the class array)
	 * @param classe a class
	 * @return a set of nodes
	 */
	private HashSet<Integer> getNodesFromClass(int classe) {
		HashSet<Integer> nodes = new HashSet<Integer>();

		for (int i=0 ; i<getSize() ; ++i) {
			if(classes[i] == classe) {
				nodes.add(i);
			}
		}
		return nodes;
	}

	/**
	 * Get the image from a start set
	 * @param origin a set of nodes
	 * @param color a color of edge (blue or red)
	 * @return a set (image)
	 */
	private HashSet<Integer> getImageFromNodes(HashSet<Integer> origin, byte color) {
		HashSet<Integer> image = new HashSet<Integer>();

		for (int node : origin) {
			image.add(get(node, color));
		}
		return image;
	}

	/**
	 * Check if there is common's nodes in the two set (usefull for check the quotient relation)
	 * 
	 * @param a a set
	 * @param b another set
	 * @return true if there is no common's node, true else
	 */
	private boolean emptyIntersection(HashSet<Integer> a, HashSet<Integer> b) {
		for (int i : a) {
			if (b.contains(i)) {
				return false;
			}
		}
		return true;
	}

	/**
	 * Get the class of a set of node (the first who are assigned, or a number of a node in the set)
	 * @param set a set of nodes
	 * @return a class
	 */
	private int getClassFromSet(HashSet<Integer> set) {
		int n = 0;
		for(int node : set) {
			if (classes[node] != -1)return classes[node];
			else n = node;
		}
		return n;
	}

	/**
	 * Change the class of nodes who are in the oldClass
	 * All nodes who was in the class oldClass are now in the class newClass
	 * @param oldClass 
	 * @param newClass
	 */
	private void updateClass(int oldClass, int newClass) {
		for (int i=0 ; i<getSize() ; ++i) {
			if(classes[i] == oldClass) {
				classes[i] = newClass;
			}
		}
	}

	/**
	 * Change the class of a set of nodes
	 * If a node in this set was in a class, change the class for all nodes in this class
	 * @param set
	 * @param newClass
	 */
	private void changeClass(HashSet<Integer> set, int newClass) {
		for (int node : set) {
			if (classes[node] == -1) {
				classes[node] = newClass;
			}
			else {
				updateClass(classes[node], newClass);
			}
		}
	}

	/**
	 * Do the calculations, for a set of node and a color
	 * (try to extend the class of the set)
	 * if the stable operation cannot be check (no extend), look if there is already a class for the image set
	 * if no nodes got a class, create a new class
	 * 
	 * @param origin a set of node who are in the same class
	 * @param color a out color (blue or red)
	 * @return the image's class. -1 if there is no modification and nothing must be do
	 */
	private int performExtend(HashSet<Integer> origin, byte color) {
		HashSet<Integer> image = getImageFromNodes(origin, color);

		//System.out.println("image " + image);

		if(!emptyIntersection(origin, image)) {	//commun in image from origin : all in the same class
			int originClass = getClassFromSet(origin);
			//System.out.println("Extend class to " + originClass);
			changeClass(image, originClass);
			return originClass;
		}
		else {	//don't fusion with origin, search if class already exists
			for (int node : image) {
				if(classes[node] != -1) {	//exist
					int sizeBefore = getNodesFromClass(classes[node]).size();

					changeClass(image, classes[node]);
					//System.out.println("found existing class " + classes[node]);

					int sizeAfter = getNodesFromClass(classes[node]).size();

					if (sizeAfter - sizeBefore == 0) {
						return -1;	//no modification, no need to trait
					}
					else {
						return classes[node];	//new states, need to trait
					}
				}
			}
			//new class, all same class
			int imgClass = getClassFromSet(image);
			changeClass(image, imgClass);
			//System.out.println("new class");
			return imgClass;
		}
	}

	/**
	 * Check if there is only one class for the graph (all nodes are in the same class)
	 * @return true / false
	 */
	private boolean oneClass() {
		int classe = classes[0];

		for(int i : classes) {
			if (i != classe) return false;
		}
		return true;
	}

	/**
	 * Add classes in the list of traitment (if there is something to do on the class)
	 * 
	 * Add in the list if the size of the set have change of if the class have change)
	 * 
	 * @param classToTrait the list of class to trait
	 * @param originClass the class of the start set
	 * @param originSize the size of the start set
	 * @param newClass the class of the image
	 */
	private void addState(List<Integer> classToTrait, int originClass, int originSize, int newClass) {
		if (newClass == -1) return;
		int newClassSize = getNodesFromClass(newClass).size();

		if (newClass == originClass && newClassSize == originSize) return;
		if (!classToTrait.contains(newClass)) {
			classToTrait.add(newClass);
		}
	}

	/**
	 * Update the classes array in order to create the minimum quotient automata
	 * 
	 * Check the out edges for classe to trait. 
	 * The quotient relation is :
	 * if there is a node who is present in the start set AND in the image, then, all the nodes of the image set got the same class as the start set
	 * then, the set of elements in this start set have change, so we must recompute is out edges
	 *  
	 * @param stablePair enter states
	 */
	private void extendClasses(StablePair stablePair) {
		//System.out.println(Arrays.toString(classes));

		LinkedList<Integer> classToTrait = new LinkedList<Integer>();	//list of class we have to trait
		//in the begining, must trait the class of the stable pair
		classToTrait.add(classes[stablePair.getA()]);	//the two nodes of the stable pair got the same class

		while(classToTrait.size() > 0) {
			//originClass : class to trait
			int originClass = classToTrait.remove();
			//System.out.println("origin Class " + originClass);
			//get the nodes who are in this class
			HashSet<Integer> origin = getNodesFromClass(originClass);
			//System.out.println("Traiting : " + origin);

			//do the extend of the set with the red edge
			int newClass = performExtend(origin, REDEDGE);
			//update the list to trait
			addState(classToTrait, originClass, origin.size(), newClass);
			/*
			if(newClass == originClass) {
				origin = getNodesFromClass(newClass);
				System.out.println("update origin to " + origin);
			}
			 */
			//do the extend of the set with the blue edge
			newClass = performExtend(origin, BLUEEDGE);
			addState(classToTrait, originClass, origin.size(), newClass);

			//System.out.println("Class To Trait : " + classToTrait);

			if (oneClass()) {	//a condition of stop : all the nodes are in the same class
				break;
			}
			//System.out.println(Arrays.toString(classes));
		}
		System.out.println("  Final Classes : " + Arrays.toString(classes));
	}

	/**
	 * Browse the word in the graph, start from the node graph
	 * Word is a succession of R or B. 
	 * Get the red edge if R, blue edge if B
	 * 
	 * @param startNode start node in the graph
	 * @param word word to browse
	 * @return the final state
	 */
	public int browse(int startNode, String word) {
		int iterNode = startNode;
		for(int i=0 ; i<word.length() ; ++i) {
			switch (word.charAt(i)) {
			case 'B': {
				iterNode = getBlue(iterNode);
				break;
			}
			case 'R': {
				iterNode = getRed(iterNode);
				break;
			}
			default : {
				throw new IllegalArgumentException(word.charAt(i) + " is different of R or B in sync-word");
			}
			}
		}
		return iterNode;
	}

	/**
	 * Check if the word synchronize the graph (if browse the word from each node, they're going on the same node)
	 * @param word synchronize word to be test
	 * @return true if word synchronize, false else
	 */
	public boolean verifySyncWord(String word) {
		int endWord = browse(0, word);	//end node with start from 0
		//check if starting on other node go on the same node 
		for (int node=1 ; node<getSize() ; ++node) {
			if (browse(node, word) != endWord) {
				System.err.println(word + " isn't a syncword for this graph");
				return false;
			}
		}
		System.out.println(word + " is a syncword for this graph (" + endWord + ")");
		return true;
	}

	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		sb.append("#node : redEdge - BlueEdge\n");
		sb.append(getSize() + "\n");
		for(int i=0 ; i<n ; ++i) {
			sb.append(i + " " + red[i] + " " + blue[i] + "\n");
		}

		return sb.toString();
	}

	/**
	 * usefull for dotty
	 * @return a string in the dot syntax wich can be put in dotty program (for visualazing)
	 */
	public String toDot() {
		StringBuilder sb = new StringBuilder();

		sb.append("digraph road_coloring {\n");
		sb.append("\trankdir=LR;\n");
		sb.append("\tsize=\"" + getSize() + "\";\n");
		sb.append("\tnode [shape = circle];\n");

		for (int i=0 ; i<getSize() ; ++i) {
			sb.append("\t" + i + "-> " + getBlue(i) + " [color=\"0.650 0.700 0.700\"];\n");
			sb.append("\t" + i + "-> " + getRed(i) + " [color=\"0.002 0.999 0.999\"];\n");
		}
		sb.append("}");

		return sb.toString();
	}

	/**
	 * launch dotty and print the graph
	 */
	public void dotty() {
		try {
			String fileName = "/tmp/graph";
			BufferedWriter br = new BufferedWriter(new FileWriter(fileName));

			br.write(toDot());
			//br.flush();
			br.close();
			String[] args = {"dotty", fileName};
			Runtime.getRuntime().exec(args);
		} catch (IOException e1) {
			e1.printStackTrace();
		}
	}
}
