//    This file is part of RoadColoringImpl.
//
//    Copyright (C) 2008 Florian SIKORA sikora@univ-mlv.fr
//    http://igm.univ-mlv.fr/~fsikora/index.php?p=r
//
//    RoadColoringImpl is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    RoadColoringImpl is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    The complete GNU General Public Licence Notice can be found as the
//    `COPYING' file in the root directory.


package fr.umlv.roadcoloring.graph;

import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
/**
 * Manage the quotient graph created in each iteration 
 * The quotient graph is like the initial graph but collapses nodes with the same class
 * 
 * 
 * @author fsikora
 *
 */

public class QuotientGraph extends Graph {
	private final HashSet<Integer> differentsElements = new HashSet<Integer>();
	private final int quotientSize;

	private int maxNode;	//number of the maxNode (p), node who is the most far from a circle
	private final LinkedList<Integer> maxNodes = new LinkedList<Integer>();	//a list of maxNode candidates
	private int maxLevel = 0;	//level of maxNode
	private int circlesSize = 0;

	private NodeData[] data;	//information for each node

	//color for deep browsing
	private final static byte GREEN = 0;
	private final static byte ORANGE = 1;
	private final static byte RED = 2;
	private byte[] color;

	private final Graph originalGraph;

	private final HashSet<Integer> ignoreNodes = new HashSet<Integer>();	//node ignored (are in a class different from itself)

	/**
	 * Construct the quotient graph from the initial graph (reduce nodes)
	 * 
	 * @param g	initial graph
	 */
	public QuotientGraph(Graph g) {	
		super(g);

		this.originalGraph = g;

		for (int i=0 ; i<getSize() ; ++i) {
			/*int classI = g.getClasses()[i];

			if (classI == -1) classI = i;
			 */
			//int classI = getClassOfNode(i);
			int classI = g.getNodeClass(i);
			if (classI != i) {	//reduce to classI, ignore i
				ignoreNodes.add(i);
			}

			/*
			 *the out edges went to the class of the arrival node
			 *for exemple : if class(2) = 1 and class(4) = 3
			 *if there is a edge 1->4
			 *we put a edge 2->3
			 */
			//setBlue(i, getClassOfNode(g.getBlue(classI)));
			setBlue(i, g.getNodeClass(g.getBlue(classI)));
			//setRed(i, getClassOfNode(g.getRed(classI)));
			setRed(i, g.getNodeClass(g.getRed(classI)));
			differentsElements.add(classI);
		}

		quotientSize = differentsElements.size();
		System.out.println("  new quotient size = " + quotientSize);
	}

	/**
	 * get the class of a node from the initial graph
	 * -1 is the status of a node not in a class (so return herself)
	 * @param node node which we seach class
	 * @return the node's class
	 */
	/*
	private int getClassOfNode(int node) {
		int cl = originalGraph.getClasses()[node];
		//-1 : not initialised, node in his class
		if(cl == -1) return node;
		return cl;
	}
	 */

	/**
	 * Number of node in the quotient graph
	 * @return size
	 */
	public int getQuotientSize() {
		return quotientSize;
	}

	/*
	protected HashSet<Integer> getIgnoreNodes() {
		return ignoreNodes;
	}
	*/

	/**
	 * flip the edge of node
	 * @param node node to be flip
	 */
	protected void edgeSwap(int node) {
		super.edgeSwap(node);
		/*	System.out.println(originalGraph);
		originalGraph.edgeSwap(node);
		System.out.println(originalGraph);
		 */
	}

	/**
	 * find and return a stable pair in the graph
	 * compute the distance of each nodes to circles
	 * made a threatman exchange if necessary for finding a stable pair
	 * 
	 * @return a stable pair (2 nodes)
	 */
	protected StablePair findStablePair() {
		//compute levels (distance from circles)
		computeIndexes();
		//if the maxNodes are on the same tree, no exchange is needed, we got the stable pair
		if (maxLevel > 0 && maxNodesSameRoot()) {
			System.out.println("  No Trahtman exchange needed !");
			return computeStablePair(); 
		}
		System.out.println("  Trahtman exchanges");
		System.out.println("  p = " + maxNode);
		PrevNodes prevNodes = new PrevNodes(this, ignoreNodes);

		int u = -1;
		//System.out.println("preds of p : " + prevNodes.getAllPred(maxNode, BLUEEDGE));

		if (maxLevel == 0) {
			while ((u = prevNodes.getOnePred(maxNode, BLUEEDGE)) == -1) {	//case all nodes on circle, can have no pred. with u circle -> choose another one
				maxNode = maxNodes.remove();
			}
		}
		//case where p == u, so u isn't on the circle, but after the exchange, theren't level increase
		else if(maxLevel > 0) {
			for(int candidateU : prevNodes.getAllPred(maxNode, BLUEEDGE)) {
				if (candidateU != maxNode) {
					u = candidateU;
					break;
				}
			}
		}
		if (u == -1) throw new AssertionError("Cannot find a u (blue predecessor of p) different of p !");
		System.out.println("  u = " + u);

		return trahtmanExchanges(u, prevNodes);
		/*
		int previousCircleSize = data[maxNode].getCircle().getSize();
		int previousMaxNode = maxNode;
		System.out.println("size av " + previousCircleSize);

		if (trahtmanExchanges(u, prevNodes)) {
			//update datas
			computeIndexes();
			int newCircleSize = data[previousMaxNode].getCircle().getSize();
			System.out.println("size ap " + newCircleSize);
			if (maxLevel == 0 || !maxNodesSameRoot()) {	
				if (newCircleSize - previousCircleSize > 0) {	//increase circle size
					System.out.println("Increasing circle size");
					return null;
				}
				throw new AssertionError("Thretman Exchange failed !!");
			}
			return computeStablePair();
		}
		return null;
		 */
	}

	// TODO	a faire en O(1)
	/**
	 * maxNode must be computed
	 * the stable pair is 
	 *  - the last node in tree of MaxNode
	 *  - the node before the root in the circle
	 * @return a stable pair
	 */
	private StablePair computeStablePair() {
		int l = data[maxNode].getLevel();
		int circleSize = data[maxNode].getCircle().getSize();

		int s = getSucessorByRed(maxNode, l-1);
		int q = getSucessorByRed(s, circleSize);

		StablePair stablePair = new StablePair(q, s);
		System.out.println("  Stable Pair : " + stablePair);

		return stablePair;
	}

	/**
	 * Compute k2
	 * u lead to v by red.
	 * k2 is distance from v to root r
	 * 
	 * @param u, predcesor of p by blue, u is on the circle
	 * @return k2
	 */
	private int computeK2(int u) {
		//u is on the circle
		int k2 = 0;
		int seek = getRed(u);
		while (seek != data[maxNode].getRootAbsolute()) {
			seek = getRed(seek);
			k2++;
		}
		return k2;
	}

	/**
	 * Performs the trahtman exchanges
	 * exchange the correct node function of the case 
	 * Call only if there is node with max level on different tree
	 * At the end of exchange, there is a new maxNode with a level higher, or the red circle size as increase
	 * then, can find a stable pair (or reperform algorithm with the new coloring) 
	 *  
	 * @param u the predecesor by blue of maxNode
	 * @param prevNodes class manage all the predecessor of the graph
	 * @return null if circle size increase, the stable pair else
	 */
	//private boolean trahtmanExchanges(int u, PrevNodes prevNodes) {
	private StablePair trahtmanExchanges(int u, PrevNodes prevNodes) {
		//int previousCircleSize = data[maxNode].getCircle().getSize();
		int previousCirclesSize = circlesSize;
		//int previousMaxNode = maxNode;
		//System.out.println("size av " + previousCircleSize);
		//System.out.println("size av " + circlesSize);

		int l = data[maxNode].getLevel();
		if (l == 0) {	//case 1
			//all reds are on cycle. Must fund a red node where his blue edge go on a other node than the red edge
			while(getBlue(maxNode) == getRed(maxNode)) {
				System.out.println("   Blue edge of p = " + maxNode + "don't get out of the circle. Choose another node on the circle");
				maxNode = getRed(maxNode);
				System.out.println("   New p is " + maxNode);
			}
			System.out.println("   Case 1 : all red nodes are on circles");
			System.out.println("   exchange edge of p = " + maxNode);
			System.out.println("   new max is q, next by blue after exchange");
			edgeSwap(maxNode);
		}
		else if (data[u].getLevel() == 0 && data[u].getCircle().getCircleNumber() == data[maxNode].getCircle().getCircleNumber()) {	//u is on the circle
			System.out.println("   u is on the circle");
			int k2 = computeK2(u);

			System.out.println("   k2 = " + k2);
			System.out.println("   l = " + l);

			if (k2 > l) {	//case 3, k2 > l
				System.out.println("   case 3 :  k2 > l");
				System.out.println("   exchange u edge");
				edgeSwap(u);
				System.out.println("   new maximal is ancestors by red of v (next of u by blue after exchange)");
			}
			else if (k2 == l) {
				System.out.println("   case k2 == l");

				int r = data[maxNode].getRootAbsolute();	//root of the tree
				List<Integer> prevR = prevNodes.getAllPred(r, REDEDGE);
				//must find s, predecesor of r in the circle
				int s = 0;
				for (int prev : prevR) {
					System.out.println(data[prev]);
					if (data[prev].getCircle().getCircleNumber() == data[r].getCircle().getCircleNumber() && data[prev].getLevel() == 0) {
						s = prev;
						break;
					}
				}
				System.out.println("   s = " + s);

				if(getBlue(s) != r) {	//case 4, edge for s to w != r
					System.out.println("   case 4");
					System.out.println("   There is a edge from s to w != r");
					System.out.println("   Exchange edge from s");
					System.out.println("   after, r not on the circle, p index increase");
					edgeSwap(s);
				}
				else {	// double bunch, s by blue and by red lead to r
					//q is predecessor by red of root, on the tree
					int q = getSucessorByRed(maxNode, maxLevel-1);
					int w = getBlue(q);
					if(w != r) {	//new by blue of w (w by red lead to r (tree)
						System.out.println("   case 5");

						System.out.println("   case 5a");
						System.out.println("   flip q = " + q);
						edgeSwap(q);
						if(data[w].getLevel() == 0) {
							System.out.println("   case 5b");
							if (data[w].getCircle() != data[r].getCircle()) {
								System.out.println("   flip u = " + u);
								edgeSwap(u);
								System.out.println("   Max is ancestor of u " + u);
							}
							else {
								System.out.println("   w on the circle same circle -> case 5d");
								System.out.println("   we are like in the case 3, with w " + w + " replacing r");
								edgeSwap(u);
							}
						}
					}
					else {
						System.out.println("   case 5c");
						System.out.println("   double bunch around the root, pair stable is q,s");
						return new StablePair(q, s);
					}
				}
			}
			else {
				//k2 < l -> red circle increase
				System.out.println("   case k2 < l");
				System.out.println("   exchange u increase circle size");
				edgeSwap(u);
				//return false;
				return null;
			}
		}
		else {	//case2, u not on circle
			System.out.println("   Case 2 : u is not on the circle");
			System.out.println("   flip u edges");
			System.out.println("   New maximal is ancestors by red of u " + u);
			edgeSwap(u);
		}

		//made the update of datas, and compute the stablePair
		computeIndexes();
		//int newCircleSize = data[previousMaxNode].getCircle().getSize();
		//System.out.println("size ap " + newCircleSize);
		//System.out.println("size ap " + circlesSize);
		if (maxLevel == 0 || !maxNodesSameRoot()) {	
			//if (newCircleSize - previousCircleSize > 0) {	//increase circle size
			if (circlesSize - previousCirclesSize > 0) {
				System.out.println("   Increasing circles size");
				return null;
			}
			throw new AssertionError("Thretman Exchange failed !!");
		}
		return computeStablePair();
	}


	/**
	 * test if all the maxNodes comes from the same root
	 * the list of maxNode is in maxNodes, member of the class, filled by computeIndexes
	 * @return true if all the maxNode lead to the same root (no exchange needed), false else
	 */
	private boolean maxNodesSameRoot() {
		int root = data[maxNodes.get(0)].getRootAbsolute();
		for (int mNode : maxNodes) {
			if (data[mNode].getRootAbsolute() != root) return false;
		}
		return true;
	}

	/**
	 * Run the computation of indexes, i.e. the levels or each nodes (distance to the circle, root of the tree)
	 * fill the data tab.
	 * fill the maxNodes list with a list of nodes with max level
	 * 
	 */
	protected void computeIndexes() {
		data = new NodeData[getSize()];
		color = new byte[getSize()];

		maxLevel = 0;
		maxNode = 0;
		circlesSize = 0;
		maxNodes.clear();

		for (int i=0 ; i<getSize() ; ++i) {
			color[i] = GREEN;
		}
		int circleNumber = 0;

		//must visit all nodes
		for (int i=0 ; i<getSize() ; ++i) {
			if(color[i] == GREEN && !ignoreNodes.contains(i)) {	//ignoreNodes : don't visit nodes who are concatained in a node (quotient graph)
				indexesExplore(i, new Circle(circleNumber));
				circleNumber++;
			}
		}

		System.out.println(" Levels :");
		for (int i=0 ; i<getSize()  ; ++i) {
			System.out.println("\tNode " + i + " : " + data[i]);
		}

	}

	/**
	 * recursive call for fill the data tab (levels of nodes)
	 * @param node current node
	 * @param circle number of current circle
	 */
	private void indexesExplore(int node, Circle circle) {
		color[node] = ORANGE;
		int next = getRed(node);

		if (color[next] == GREEN) { //next by red not visited, call recursivly
			indexesExplore(next, circle);
		}

		if(color[next] == RED) {	//next by red already visited, current node can get data from its
			if (data[next].getLevel() > 0) {	//next in the tree
				data[node] = new NodeData(data[next].getCircle(), data[next].getLevel()+1, data[next].getRootAbsolute());
			}
			else {	//on circle or first in tree
				if(data[next].getCircle().isCompleted()) {	//circle completed, new tree
					data[node] = new NodeData(data[next].getCircle(), data[next].getLevel()+1, next);
				}
				else {	//in the circle
					//data[node] = new NodeData(data[next].getCircle(), data[next].getLevel(), data[next].getRootAbsolute());
					data[node] = new NodeData(data[next].getCircle(), data[next].getLevel(), node);
				}
			}
		}

		if(color[getRed(node)] == ORANGE) {	//next by red is in visiting : the circle is completed
			circle.setRoot(next);
			data[node] = new NodeData(circle, 0, next);
		}

		if(data[node].getLevel() == maxLevel) {	//other node with same length
			maxNodes.add(node);
		}		
		else if(data[node].getLevel() > maxLevel) {	//new maxNode
			maxLevel = data[node].getLevel();
			maxNode = node;
			maxNodes.clear();
			maxNodes.add(maxNode);
		}

		if(data[node].getLevel() == 0) circlesSize++;
		circle.putNode(node);
		color[node] = RED;
	}
}
